"""
.. module:: python-aio_etcd
   :synopsis: An asynchronus python etcd client.

.. moduleauthor:: Jose Plana <jplana@gmail.com>


"""
import logging
#from http.client import HTTPException
from aiohttp.web_exceptions import HTTPException
try:
    from aiohttp.errors import DisconnectedError,ClientConnectionError,ClientResponseError
except ImportError:
    from aiohttp.client_exceptions import ServerDisconnectedError as DisconnectedError
    from aiohttp.client_exceptions import ClientConnectionError,ClientResponseError
from aiohttp.helpers import BasicAuth
import socket
import aiohttp
from urllib3.exceptions import HTTPError
from urllib3.exceptions import ReadTimeoutError
import json
import ssl
import dns.resolver
import aio_etcd as etcd
import asyncio
from functools import wraps

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


_log = logging.getLogger(__name__)


class Client(object):

    """
    Asynchronous client for etcd, the distributed log service using raft.
    """

    _MGET = 'GET'
    _MPUT = 'PUT'
    _MPOST = 'POST'
    _MDELETE = 'DELETE'
    _comparison_conditions = set(('prevValue', 'prevIndex', 'prevExist', 'refresh'))
    _read_options = set(('recursive', 'wait', 'waitIndex', 'sorted', 'quorum'))
    _del_conditions = set(('prevValue', 'prevIndex'))

    _client = None

    def __init__(
            self,
            host='127.0.0.1',
            port=2379,
            srv_domain=None,
            version_prefix='/v2',
            allow_redirect=True,
            protocol='http',
            cert=None,
            ca_cert=None,
            username=None,
            password=None,
            allow_reconnect=False,
            use_proxies=False,
            expected_cluster_id=None,
            per_host_pool_size=10,
            ssl_verify=ssl.CERT_REQUIRED,
            loop=None,
            lock_prefix="/_locks"
    ):
        """
        Initialize the client.

        Args:
            host (mixed):
                           If a string, IP to connect to.
                           If a tuple ((host, port), (host, port), ...)

            port (int):  Port used to connect to etcd.

            srv_domain (str): Domain to search the SRV record for cluster autodiscovery.

            version_prefix (str): Url or version prefix in etcd url (default=/v2).

            allow_redirect (bool): allow the client to connect to other nodes.

            protocol (str):  Protocol used to connect to etcd.

            cert (mixed):   If a string, the whole ssl client certificate;
                            if a tuple, the cert and key file names.

            ca_cert (str): The ca certificate. If pressent it will enable
                           validation.

            username (str): username for etcd authentication.

            password (str): password for etcd authentication.

            allow_reconnect (bool): allow the client to reconnect to another
                                    etcd server in the cluster in the case the
                                    default one does not respond.

            use_proxies (bool): we are using a list of proxies to which we connect,
                                 and don't want to connect to the original etcd cluster.

            expected_cluster_id (str): If a string, recorded as the expected
                                       UUID of the cluster (rather than
                                       learning it from the first request),
                                       reads will raise EtcdClusterIdChanged
                                       if they receive a response with a
                                       different cluster ID.
            per_host_pool_size (int): specifies maximum number of connections to pool
                                      by host. By default this will use up to 10
                                      connections.
            lock_prefix (str): Set the key prefix at etcd when client to lock object.
                                      By default this will be use /_locks.
        """

        # If a DNS record is provided, use it to get the hosts list
        if srv_domain is not None:
            try:
                host = self._discover(srv_domain)
            except Exception as e:
                _log.error("Could not discover the etcd hosts from %s: %s",
                           srv_domain, e)

        self._protocol = protocol
        self._loop = loop if loop is not None else asyncio.get_event_loop()

        def uri(protocol, host, port):
            return '%s://%s:%d' % (protocol, host, port)

        if not isinstance(host, tuple):
            self._machines_cache = []
            self._base_uri = uri(self._protocol, host, port)
        else:
            if not allow_reconnect:
                _log.error("List of hosts incompatible with allow_reconnect.")
                raise etcd.EtcdException("A list of hosts to connect to was given, but reconnection not allowed?")
            self._machines_cache = [uri(self._protocol, *conn) for conn in host]
            self._base_uri = self._machines_cache.pop(0)

        self.expected_cluster_id = expected_cluster_id
        self.version_prefix = version_prefix

        self._allow_redirect = allow_redirect
        self._use_proxies = use_proxies
        self._allow_reconnect = allow_reconnect
        self._lock_prefix = lock_prefix

        # SSL Client certificate support
        ssl_ctx = ssl.create_default_context()
        if protocol == 'https':
            # If we don't allow TLSv1, clients using older version of OpenSSL
            # (<1.0) won't be able to connect.
            _log.debug("HTTPS enabled.")
            #kw['ssl_version'] = ssl.PROTOCOL_TLSv1
            if ssl_verify == ssl.CERT_NONE:
                ssl_ctx.check_hostname = False
            ssl_ctx.verify_mode = ssl_verify

        if cert:
            if isinstance(cert, tuple):
                # Key and cert are separate
                ssl_ctx.load_cert_chain(*cert)
            else:
                ssl_ctx.load_cert_chain(cert)

        if ca_cert:
            ssl_ctx.load_verify_locations(ca_cert)

        self.username = None
        self.password = None
        if username and password:
            self.username = username
            self.password = password
        elif username:
            _log.warning('Username provided without password, both are required for authentication')
        elif password:
            _log.warning('Password provided without username, both are required for authentication')
        if self._allow_reconnect:
            # we need the set of servers in the cluster in order to try
            # reconnecting upon error. The cluster members will be
            # added to the hosts list you provided. If you are using
            # proxies, set all
            #
            # Beware though: if you input '127.0.0.1' as your host and
            # etcd advertises 'localhost', both will be in the
            # resulting list.

            # If we're connecting to the original cluster, we can
            # extend the list given to the client with what we get
            # from self.machines
            self._machines_available = self._use_proxies
            self._machines_cache = list(set(self._machines_cache))

            if self._base_uri in self._machines_cache:
                self._machines_cache.remove(self._base_uri)
            _log.debug("Machines cache initialised to %s",
                       self._machines_cache)
        else:
            self._machines_available = True

        if protocol == "https":
            conn = aiohttp.TCPConnector(ssl_context=ssl_ctx, loop=loop)
            self._client = aiohttp.ClientSession(connector=conn, loop=loop)
        else:
            self._client = aiohttp.ClientSession(loop=loop)

    def __del__(self):
        self.close()

    def close(self):
        """Explicitly release the etcd connection(s)."""
        if self._client is not None:
            self._client.close()
            self._client = None

    async def _update_machines(self):
        self._machines_cache = await self.machines()
        self._machines_available = True

        # Versions set to None. They will be set upon first usage.
        self._version = self._cluster_version = None

    def _set_version_info(self):
        """
        Sets the version information provided by the server.
        """
        # Set the version
        version_info = json.loads(self.http.request(
            self._MGET,
            self._base_uri + '/version',
            headers=self._get_headers(),
            timeout=self.read_timeout,
            redirect=self.allow_redirect).data.decode('utf-8'))
        self._version = version_info['etcdserver']
        self._cluster_version = version_info['etcdcluster']

    def _discover(self, domain):
        srv_name = "_etcd._tcp.{}".format(domain)
        answers = dns.resolver.query(srv_name, 'SRV')
        hosts = []
        for answer in answers:
            hosts.append(
                (answer.target.to_text(omit_final_dot=True), answer.port))
        _log.debug("Found %s", hosts)
        if not len(hosts):
            raise ValueError("The SRV record is present but no host were found")
        return tuple(hosts)

    @property
    def base_uri(self):
        """URI used by the client to connect to etcd."""
        return self._base_uri

    @property
    def host(self):
        """Node to connect to etcd."""
        return urlparse(self._base_uri).netloc.split(':')[0]

    @property
    def port(self):
        """Port to connect to etcd."""
        return int(urlparse(self._base_uri).netloc.split(':')[1])

    @property
    def protocol(self):
        """Protocol used to connect to etcd."""
        return self._protocol

    @property
    def allow_redirect(self):
        """Allow the client to connect to other nodes."""
        return self._allow_redirect

    @property
    def lock_prefix(self):
        """Get the key prefix at etcd when client to lock object."""
        return self._lock_prefix

    async def machines(self):
        """
        Members of the cluster.

        Returns:
            list. str with all the nodes in the cluster.

        >>> print client.machines
        ['http://127.0.0.1:4001', 'http://127.0.0.1:4002']
        """
        # We can't use api_execute here, or it causes a logical loop
        retries = 0
        while retries < 5:
            for m in [self._base_uri]+self._machines_cache:
                try:
                    uri = m + self.version_prefix + '/machines'
                    response = await self._client.request(
                        self._MGET,
                        uri,
                        allow_redirects=self.allow_redirect,
                    )

                    response = await self._handle_server_response(response)
                    response = await response.read()
                    if response != b"":
                        machines = [
                            node.strip() for node in response.decode('utf-8').split(',')
                        ]
                        _log.debug("Retrieved list of machines: %s", machines)
                        return machines
                except (HTTPException, socket.error, DisconnectedError, ClientConnectionError) as e:
                    # We can't get the list of machines, if one server is in the
                    # machines cache, try on it
                    _log.error("Failed to get list of machines from %s%s: %r",
                            self._base_uri, self.version_prefix, e)
            retries += 1
            await asyncio.sleep(retries/10, loop=self._loop)
        raise etcd.EtcdException("Could not get the list of servers, "
                                    "maybe you provided the wrong "
                                    "host(s) to connect to?")

    async def members(self):
        """
        A more structured view of peers in the cluster.
        """
        # Empty the members list
        self._members = {}
        try:
            response = await self.api_execute(self.version_prefix + '/members', self._MGET)
            data = await response.read()
            res = json.loads(data.decode('utf-8'))
            for member in res['members']:
                self._members[member['id']] = member
            return self._members
        except Exception as e:
            raise etcd.EtcdException("Could not get the members list, maybe the cluster has gone away?") from e

    async def leader(self):
        """
        Returns:
            dict. the leader of the cluster.

        >>> print (loop.run_until_complete(client.leader()))
        {"id":"ce2a822cea30bfca","name":"default","peerURLs":["http://localhost:2380","http://localhost:7001"],"clientURLs":["http://127.0.0.1:4001"]}
        """
        try:
            response = await self.api_execute(self.version_prefix + '/stats/self', self._MGET)
            data = await response.read()
            leader = json.loads(data.decode('utf-8'))
            return (await self.members())[leader['leaderInfo']['leader']]
        except Exception as exc:
            raise etcd.EtcdException("Cannot get leader data") from exc

    def stats(self):
        """
        Returns:
            dict. the stats of the local server
        """
        return self._stats()
    stats._is_coroutine = True

    def leader_stats(self):
        """
        Returns:
            dict. the stats of the leader
        """
        return self._stats('leader')
    leader_stats._is_coroutine = True

    def store_stats(self):
        """
        Returns:
           dict. the stats of the kv store
        """
        return self._stats('store')
    store_stats._is_coroutine = True

    async def _stats(self, what='self'):
        """ Internal method to access the stats endpoints"""
        data = await self.api_execute(self.version_prefix + '/stats/' + what, self._MGET)
        data = await data.read()
        data = data.decode('utf-8')
        try:
            return json.loads(data)
        except (TypeError,ValueError) as e:
            raise etcd.EtcdException("Cannot parse json data in the response") from e

    @property
    def version(self):
        """
        Version of etcd.
        """
        if not self._version:
            self._set_version_info()
        return self._version

    @property
    def cluster_version(self):
        """
        Version of the etcd cluster.
        """
        if not self._cluster_version:
            self._set_version_info()

        return self._cluster_version

    @property
    def key_endpoint(self):
        """
        REST key endpoint.
        """
        return self.version_prefix + '/keys'

    async def contains(self, key):
        """
        Check if a key is available in the cluster.

        >>> print 'key' in client
        True
        """
        try:
            await self.get(key)
            return True
        except etcd.EtcdKeyNotFound:
            return False

    def _sanitize_key(self, key):
        if not key.startswith('/'):
            key = "/{}".format(key)
        return key

    async def write(self, key, value, ttl=None, dir=False, append=False, **kwdargs):
        """
        Writes the value for a key, possibly doing atomic Compare-and-Swap

        Args:
            key (str):  Key.

            value (object):  value to set

            ttl (int):  Time in seconds of expiration (optional).

            dir (bool): Set to true if we are writing a directory; default is false.

            append (bool): If true, it will post to append the new value to the dir, creating a sequential key. Defaults to false.

            Other parameters modifying the write method are accepted:


            prevValue (str): compare key to this value, and swap only if corresponding (optional).

            prevIndex (int): modify key only if actual modifiedIndex matches the provided one (optional).

            prevExist (bool): If false, only create key; if true, only update key.

            refresh (bool): since 2.3.0, If true, only update the ttl, prev key must existed(prevExist=True).

        Returns:
            client.EtcdResult

        >>> print client.write('/key', 'newValue', ttl=60, prevExist=False).value
        'newValue'

        """
        _log.debug("Writing %s to key %s ttl=%s dir=%s append=%s %s",
                  value, key, ttl, dir, append, kwdargs)
        key = self._sanitize_key(key)
        params = {}
        if value is not None:
            params['value'] = value

        if ttl is not None:
            params['ttl'] = ttl

        if dir:
            if value:
                raise etcd.EtcdException(
                    'Cannot create a directory with a value')
            params['dir'] = "true"

        for (k, v) in kwdargs.items():
            if k in self._comparison_conditions:
                if type(v) == bool:
                    params[k] = v and "true" or "false"
                else:
                    params[k] = v

        method = append and self._MPOST or self._MPUT
        if '_endpoint' in kwdargs:
            path = kwdargs['_endpoint'] + key
        else:
            path = self.key_endpoint + key

        response = await self.api_execute(path, method, params=params)
        return (await self._result_from_response(response))

    def refresh(self, key, ttl, **kwdargs):
        """
        (Since 2.3.0) Refresh the ttl of a key without notifying watchers.

        Keys in etcd can be refreshed without notifying watchers,
        this can be achieved by setting the refresh to true when updating a TTL

        You cannot update the value of a key when refreshing it

        @see: https://github.com/coreos/etcd/blob/release-2.3/Documentation/api.md#refreshing-key-ttl

        Args:
            key (str):  Key.

            ttl (int):  Time in seconds of expiration (optional).

            Other parameters modifying the write method are accepted as `EtcdClient.write`.
        """
        # overwrite kwdargs' prevExist
        kwdargs['prevExist'] = True
        return self.write(key=key, value=None, ttl=ttl, refresh=True, **kwdargs)

    def update(self, obj):
        """
        Updates the value for a key atomically. Typical usage would be:

        c = aio_etcd.Client()
        o = await c.read("/somekey")
        o.value += 1
        await c.update(o)

        Args:
            obj (aio_etcd.EtcdResult):  The object that needs updating.

        This method returns a coroutine.
        """
        _log.debug("Updating %s to %s.", obj.key, obj.value)
        kwdargs = {
            'dir': obj.dir,
            'ttl': obj.ttl,
            'prevExist': True
            }

        if not obj.dir:
            # prevIndex on a dir causes a 'not a file' error. d'oh!
            kwdargs['prevIndex'] = obj.modifiedIndex
        return self.write(obj.key, obj.value, **kwdargs)

    async def read(self, key, **kwdargs):
        """
        Returns the value of the key 'key'.

        Args:
            key (str):  Key.

            Recognized kwd args

            recursive (bool): If you should fetch recursively a dir

            wait (bool): If we should wait and return next time the key is changed

            waitIndex (int): The index to fetch results from.

            sorted (bool): Sort the output keys (alphanumerically)

        Returns:
            client.EtcdResult (or an array of client.EtcdResult if a
            subtree is queried)

        Raises:
            KeyValue:  If the key doesn't exists.

        >>> print client.get('/key').value
        'value'

        """
        _log.debug("Issuing read for key %s with args %s", key, kwdargs)
        key = self._sanitize_key(key)

        params = {}
        for (k, v) in kwdargs.items():
            if k in self._read_options:
                if type(v) == bool:
                    params[k] = v and "true" or "false"
                elif v is not None:
                    params[k] = v

        response = await self.api_execute(
            self.key_endpoint + key, self._MGET, params=params)
        return (await self._result_from_response(response))

    async def delete(self, key, recursive=None, dir=None, **kwdargs):
        """
        Removed a key from etcd.

        Args:

            key (str):  Key.

            recursive (bool): if we want to recursively delete a directory, set
                              it to true

            dir (bool): if we want to delete a directory, set it to true

            prevValue (str): compare key to this value, and swap only if
                             corresponding (optional).

            prevIndex (int): modify key only if actual modifiedIndex matches the
                             provided one (optional).

        Returns:
            client.EtcdResult

        Raises:
            KeyValue:  If the key doesn't exists.

        >>> print client.delete('/key').key
        '/key'

        """
        _log.debug("Deleting %s recursive=%s dir=%s extra args=%s",
                   key, recursive, dir, kwdargs)
        key = self._sanitize_key(key)

        kwds = {}
        if recursive is not None:
            kwds['recursive'] = recursive and "true" or "false"
        if dir is not None:
            kwds['dir'] = dir and "true" or "false"

        for k in self._del_conditions:
            if k in kwdargs:
                kwds[k] = kwdargs[k]
        _log.debug("Calculated params = %s", kwds)

        response = await self.api_execute(
            self.key_endpoint + key, self._MDELETE, params=kwds)
        return (await self._result_from_response(response))

    def pop(self, key, recursive=None, dir=None, **kwdargs):
        """
        Remove specified key from etcd and return the corresponding value.

        Args:

            key (str):  Key.

            recursive (bool): if we want to recursively delete a directory, set
                              it to true

            dir (bool): if we want to delete a directory, set it to true

            prevValue (str): compare key to this value, and swap only if
                             corresponding (optional).

            prevIndex (int): modify key only if actual modifiedIndex matches the
                             provided one (optional).

        Returns:
            A coroutine returning client.EtcdResult

        Raises:
            KeyValue:  If the key doesn't exists.

        >>> print client.pop('/key').value
        'value'

        """
        return self.delete(key=key, recursive=recursive, dir=dir, **kwdargs)._prev_node
    pop._is_coroutine = True

    # Higher-level methods on top of the basic primitives
    def test_and_set(self, key, value, prev_value, ttl=None):
        """
        Atomic test & set operation.
        It will check if the value of 'key' is 'prev_value',
        if the the check is correct will change the value for 'key' to 'value'
        if the the check is false an exception will be raised.

        Args:
            key (str):  Key.
            value (object):  value to set
            prev_value (object):  previous value.
            ttl (int):  Time in seconds of expiration (optional).

        Returns:
            A coroutine returning client.EtcdResult

        Raises:
            ValueError: When the 'prev_value' is not the current value.

        >>> print client.test_and_set('/key', 'new', 'old', ttl=60).value
        'new'

        """
        return self.write(key, value, prevValue=prev_value, ttl=ttl)
    test_and_set._is_coroutine = True

    def set(self, key, value, ttl=None):
        """
        Compatibility: sets the value of the key 'key' to the value 'value'

        Args:
            key (str):  Key.
            value (object):  value to set
            ttl (int):  Time in seconds of expiration (optional).

        Returns:
            A coroutine returning client.EtcdResult

        Raises:
           etcd.EtcdException: when something weird goes wrong.

        """
        return self.write(key, value, ttl=ttl)
    set._is_coroutine = True

    def get(self, key):
        """
        Returns the value of the key 'key'.

        Args:
            key (str):  Key.

        Returns:
            A coroutine returning client.EtcdResult

        Raises:
            KeyError:  If the key doesn't exists.

        >>> print client.get('/key').value
        'value'

        """
        return self.read(key)
    get._is_coroutine = True

    async def watch(self, key, index=None, recursive=None):
        """
        Blocks until a new event has been received, starting at index 'index'

        Args:
            key (str):  Key.

            index (int): Index to start from.

            timeout (int):  max seconds to wait for a read.

        Returns:
            A coroutine returning client.EtcdResult

        Raises:
            KeyValue:  If the key doesn't exist.

        >>> print client.watch('/key').value
        'value'

        """
        _log.debug("Wait %s on %s", index, key)
        if index:
            res = await self.read(key, wait=True, waitIndex=index, recursive=recursive)
        else:
            res = await self.read(key, wait=True, recursive=recursive)
        _log.debug("Wait %s on %s done: %s",index, key, res)
        return res

    async def eternal_watch(self, key, callback, index=None, recursive=None):
        """
        Generator that will call the callback every time a key changes.
        Note that this method will block forever until an event is generated
        and the callback function raises StopWatching.

        Args:
            key (str):  Key to subcribe to.
            index (int):  Index from where the changes will be received.

        Returns:
            Index to continue watching from

        >>> for event in client.eternal_watch('/subcription_key'):
        ...     print event.value
        ...
        value1
        value2

        """
        local_index = index
        while True:
            response = await self.watch(key, index=local_index, recursive=recursive)
            local_index = response.modifiedIndex + 1
            try:
                res = callback(response)
                if isinstance(res, asyncio.Future) or asyncio.iscoroutine(res):
                    await res
            except etcd.StopWatching:
                return local_index

    def get_lock(self, *args, **kwargs):
        raise NotImplementedError('Lock primitives were removed from etcd 2.0')

    @property
    def election(self):
        raise NotImplementedError('Election primitives were removed from etcd 2.0')

    async def _result_from_response(self, response):
        """ Creates an EtcdResult from json dictionary """
        raw_response = await response.read()
        try:
            res = await response.json()
        except (TypeError, ValueError, UnicodeError) as e:
            raise etcd.EtcdException(
                'Server response was not valid JSON: %r', raw_response)
        _log.debug("result: %s", res)
        try:
            r = etcd.EtcdResult(**res)
            if response.status == 201:
                r.newKey = True
            r.parse_headers(response)
            return r
        except Exception as e:
            raise etcd.EtcdException(
                'Unable to decode server response') from e

    def _next_server(self, cause=None):
        """ Selects the next server in the list, refreshes the server list. """
        _log.debug("Selection next machine in cache. Available machines: %s",
                   self._machines_cache)
        try:
            mach = self._machines_cache.pop()
        except IndexError as e:
            _log.error("Machines cache is empty, no machines to try.")
            raise etcd.EtcdConnectionFailed('No more machines in the cluster', cause=cause) from e
        else:
            _log.info("Selected new etcd server %s", mach)
            return mach

    def _wrap_request(payload):
        @wraps(payload)
        async def wrapper(self, path, method, params=None):
            some_request_failed = False
            response = False

            if not path.startswith('/'):
                raise ValueError('Path does not start with /')

            if not self._machines_available:
                await self._update_machines()

            while not response:
                some_request_failed = False
                try:
                    response = await payload(self, path, method, params=params)
                    # Check the cluster ID hasn't changed under us.  We use
                    # preload_content=False above so we can read the headers
                    # before we wait for the content of a watch.
                    self._check_cluster_id(response)
                    # Now force the data to be preloaded in order to trigger any
                    # IO-related errors in this method rather than when we try to
                    # access it later.
                    _ = await response.read()
                    # urllib3 doesn't wrap all httplib exceptions and earlier versions
                    # don't wrap socket errors either.
                except (ClientResponseError, DisconnectedError, HTTPException, socket.error) as e:
                    if (isinstance(params, dict) and
                        params.get("wait") == "true" and
                        isinstance(e, ReadTimeoutError)):
                        _log.debug("Watch timed out.")
                        raise etcd.EtcdWatchTimedOut(
                            "Watch timed out: %r" % e,
                            cause=e
                        )
                    _log.error("Request to server %s failed: %r",
                               self._base_uri, e)
                    if self._allow_reconnect:
                        _log.info("Reconnection allowed, looking for another "
                                  "server.")
                        # _next_server() raises EtcdException if there are no
                        # machines left to try, breaking out of the loop.
                        self._base_uri = self._next_server(cause=e)
                        some_request_failed = True

                        # if exception is raised on _ = response.data
                        # the condition for while loop will be False
                        # but we should retry
                        response = False
                    else:
                        _log.debug("Reconnection disabled, giving up.")
                        raise etcd.EtcdConnectionFailed(
                            "Connection to etcd failed due to %r" % e,
                            cause=e
                        )
                except etcd.EtcdClusterIdChanged as e:
                    _log.warning(e)
                    raise
                except asyncio.CancelledError:
                    # don't complain
                    raise
                except Exception:
                    _log.exception("Unexpected request failure, re-raising.")
                    raise
                else:
                    try:
                        response = await self._handle_server_response(response)
                    except etcd.EtcdException as e:
                        if "during rolling upgrades" in e.payload['message']:
                            response = False
                            some_request_failed = True
                        else:
                            raise

                if some_request_failed:
                    if not self._use_proxies:
                        # The cluster may have changed since last invocation
                        self._machines_cache = await self.machines()
                    if self._base_uri in self._machines_cache:
                        self._machines_cache.remove(self._base_uri)
            return response
        return wrapper

    @_wrap_request
    def api_execute(self, path, method, params=None):
        """ Executes the query. """

        if not path.startswith('/'):
            raise ValueError('Path does not start with /')

        url = self._base_uri + path

        return self._client.request(
            method,
            url,
            params=params,
            auth=self._get_auth(),
            allow_redirects=self.allow_redirect,
            )

    @_wrap_request
    def api_execute_json(self, path, method, params=None):
        url = self._base_uri + path
        json_payload = json.dumps(params)
        headers = { 'Content-Type': 'application/json' }
        return self._client.request(method,
                                    url,
                                    data=json_payload,
                                    allow_redirects=self.allow_redirect,
                                    auth=self._get_auth(),
                                    headers=headers)

    def _check_cluster_id(self, response):
        cluster_id = response.headers.get("x-etcd-cluster-id", None)
        if not cluster_id:
            # _log.warning("etcd response did not contain a cluster ID")
            return
        id_changed = (self.expected_cluster_id and
                      cluster_id != self.expected_cluster_id)
        # Update the ID so we only raise the exception once.
        old_expected_cluster_id = self.expected_cluster_id
        self.expected_cluster_id = cluster_id
        if id_changed:
            # Defensive: clear the pool so that we connect afresh next
            # time.
            raise etcd.EtcdClusterIdChanged(
                'The UUID of the cluster changed from {} to '
                '{}.'.format(old_expected_cluster_id, cluster_id))

    async def _handle_server_response(self, response):
        """ Handles the server response """
        if response.status in [200, 201]:
            return response

        else:
            data = await response.read()
            resp = data.decode('utf-8')

            # throw the appropriate exception
            try:
                r = json.loads(resp)
                r['status'] = response.status
            except (TypeError, ValueError):
                # Bad JSON, make a response locally.
                r = {"message": "Bad response",
                     "cause": str(resp)}
            etcd.EtcdError.handle(r)

    def _get_auth(self):
        if self.username and self.password:
            return BasicAuth(self.username, self.password)
        return None




import logging
from etcd import *
from .client import Client
from .lock import Lock

__VERSION__ = (0,4,5,0)

_log = logging.getLogger(__name__)

class StopWatching(BaseException):
    pass

_EtcdResult = EtcdResult
class EtcdResult(_EtcdResult):
    def parse_headers(self, response):
        headers = response.headers
        self.etcd_index = int(headers.get('x-etcd-index', 1))
        self.raft_index = int(headers.get('x-raft-index', 1))

    def get_subtree(self, leaves_only=False):
        """
        Get all the subtree resulting from a recursive=true call to etcd.

        Args:
            leaves_only (bool): if true, only value nodes are returned


        """
        if not self._children:
            #if the current result is a leaf, return itself
            yield self
            return
        else:
            # node is not a leaf
            if not leaves_only:
                yield self
            for n in self._children:
                node = EtcdResult(None, n)
                for child in node.get_subtree(leaves_only=leaves_only):
                    yield child
        return

    @property
    def leaves(self):
        return self.get_subtree(leaves_only=True)

    @property
    def children(self):
        """ Deprecated, use EtcdResult.leaves instead """
        return self.leaves

    def __eq__(self, other):
        if not (type(self) is type(other)):
            return False
        for k in self._node_props.keys():
            try:
                a = getattr(self, k)
                b = getattr(other, k)
                if a != b:
                    return False
            except:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class EtcdException(Exception):

    """
    Generic Etcd Exception.
    """
    def __init__(self, message=None, payload=None):
        super(EtcdException, self).__init__(message)
        self.payload = payload


class EtcdValueError(EtcdException, ValueError):
    """
    Base class for Etcd value-related errors.
    """
    pass


class EtcdCompareFailed(EtcdValueError):
    """
    Compare-and-swap failure
    """
    pass


class EtcdClusterIdChanged(EtcdException):
    """
    The etcd cluster ID changed.  This may indicate the cluster was replaced
    with a backup.  Raised to prevent waiting on an etcd_index that was only
    valid on the old cluster.
    """
    pass


class EtcdKeyError(EtcdException):
    """
    Etcd Generic KeyError Exception
    """
    pass


class EtcdKeyNotFound(EtcdKeyError):
    """
    Etcd key not found exception (100)
    """
    pass


class EtcdNotFile(EtcdKeyError):
    """
    Etcd not a file exception (102)
    """
    pass


class EtcdNotDir(EtcdKeyError):
    """
    Etcd not a directory exception (104)
    """
    pass


class EtcdAlreadyExist(EtcdKeyError):
    """
    Etcd already exist exception (105)
    """
    pass


class EtcdEventIndexCleared(EtcdException):
    """
    Etcd event index is outdated and cleared exception (401)
    """
    pass


class EtcdConnectionFailed(EtcdException):
    """
    Connection to etcd failed.
    """
    def __init__(self, message=None, payload=None, cause=None):
        super(EtcdConnectionFailed, self).__init__(message=message,
                                                   payload=payload)
        self.cause = cause


class EtcdInsufficientPermissions(EtcdException):
    """
    Request failed because of insufficient permissions.
    """
    pass


class EtcdWatchTimedOut(EtcdConnectionFailed):
    """
    A watch timed out without returning a result.
    """
    pass


class EtcdWatcherCleared(EtcdException):
    """
    Watcher is cleared due to etcd recovery.
    """
    pass


class EtcdLeaderElectionInProgress(EtcdException):
    """
    Request failed due to in-progress leader election.
    """
    pass


class EtcdRootReadOnly(EtcdKeyError):
    """
    Operation is not valid on the root, which is read only.
    """
    pass


class EtcdDirNotEmpty(EtcdValueError):
    """
    Directory not empty.
    """
    pass


class EtcdLockExpired(EtcdException):
    """
    Our lock apparently expired while we were trying to acquire it.
    """
    pass


class EtcdError(object):
    # See https://github.com/coreos/etcd/blob/master/Documentation/v2/errorcode.md
    error_exceptions = {
        100: EtcdKeyNotFound,
        101: EtcdCompareFailed,
        102: EtcdNotFile,
        # 103: Non-public: no more peers.
        104: EtcdNotDir,
        105: EtcdAlreadyExist,
        # 106: Non-public: key is preserved.
        107: EtcdRootReadOnly,
        108: EtcdDirNotEmpty,
        # 109: Non-public: existing peer addr.
        110: EtcdInsufficientPermissions,

        200: EtcdValueError,  # Not part of v2
        201: EtcdValueError,
        202: EtcdValueError,
        203: EtcdValueError,
        204: EtcdValueError,
        205: EtcdValueError,
        206: EtcdValueError,
        207: EtcdValueError,
        208: EtcdValueError,
        209: EtcdValueError,
        210: EtcdValueError,

        # 300: Non-public: Raft internal error.
        301: EtcdLeaderElectionInProgress,

        400: EtcdWatcherCleared,
        401: EtcdEventIndexCleared,
    }

    @classmethod
    def handle(cls, payload):
        """
        Decodes the error and throws the appropriate error message

        :param payload: The decoded JSON error payload as a dict.
        """
        error_code = payload.get("errorCode")
        message = payload.get("message")
        cause = payload.get("cause")
        msg = '{} : {}'.format(message, cause)
        status = payload.get("status")
        # Some general status handling, as
        # not all endpoints return coherent error messages
        if status == 404:
            error_code = 100
        elif status == 401:
            error_code = 110
        exc = cls.error_exceptions.get(error_code, EtcdException)
        if issubclass(exc, EtcdException):
            raise exc(msg, payload)
        else:
            raise exc(msg)


# Attempt to enable urllib3's SNI support, if possible
# Blatantly copied from requests.
try:
    from urllib3.contrib import pyopenssl
    pyopenssl.inject_into_urllib3()
except ImportError:
    pass



import logging
import aio_etcd
import uuid
import asyncio

_log = logging.getLogger(__name__)

class Lock(object):
    """
    Locking recipe for etcd, inspired by the kazoo recipe for zookeeper
    """

    def __init__(self, client, lock_name):
        self.client = client
        self.name = lock_name
        # props to Netflix Curator for this trick. It is possible for our
        # create request to succeed on the server, but for a failure to
        # prevent us from getting back the full path name. We prefix our
        # lock name with a uuid and can check for its presence on retry.
        self._uuid = uuid.uuid4().hex
        self.path = "{}/{}".format(client.lock_prefix, lock_name) 
        self.is_taken = False
        self._sequence = None
        _log.debug("Initiating lock for %s with uuid %s", self.path, self._uuid)

    @property
    def uuid(self):
        """
        The unique id of the lock
        """
        return self._uuid

    async def set_uuid(self, value):
        old_uuid = self._uuid
        self._uuid = value
        if not (await self._find_lock()):
            _log.warn("The hand-set uuid was not found, refusing")
            self._uuid = old_uuid
            raise ValueError("Nonexistent UUID")

    async def is_acquired(self):
        """
        tells us if the lock is acquired
        """
        if not self.is_taken:
            _log.debug("Lock not taken")
            return False
        try:
            await self.client.read(self.lock_key)
            return True
        except aio_etcd.EtcdKeyNotFound:
            _log.warn("Lock was supposedly taken, but we cannot find it")
            self.is_taken = False
            return False

    async def acquire(self, blocking=True, lock_ttl=3600):
        """
        Acquire the lock.

        :param blocking Block until the lock is obtained, or timeout is reached
        :param lock_ttl The duration of the lock we acquired, set to None for eternal locks
        :param timeout The time to wait before giving up on getting a lock
        """
        # First of all try to write, if our lock is not present.
        if not (await self._find_lock()):
            _log.debug("Lock not found, writing it to %s", self.path)
            res = await self.client.write(self.path, self.uuid, ttl=lock_ttl, append=True)
            self._set_sequence(res.key)
            _log.debug("Lock key %s written, sequence is %s", res.key, self._sequence)
        elif lock_ttl:
            # Renew our lock if already here!
            await self.client.write(self.lock_key, self.uuid, ttl=lock_ttl)

        # now get the owner of the lock, and the next lowest sequence
        return self._acquired(blocking=blocking)

    async def release(self):
        """
        Release the lock
        """
        if not self._sequence:
            await self._find_lock()
        try:
            _log.debug("Releasing existing lock %s", self.lock_key)
            await self.client.delete(self.lock_key)
        except aio_etcd.EtcdKeyNotFound:
            _log.info("Lock %s not found, nothing to release", self.lock_key)
            pass
        finally:
            self.is_taken = False

    async def __aenter__(self):
        """
        You can use the lock as a contextmanager
        """
        await self.acquire(blocking=True, lock_ttl=0)
        self.acquire(blocking=True, lock_ttl=None)
        return self

    async def __aexit__(self, type, value, traceback):
        await self.release()
        return False

    async def _acquired(self, blocking=True):
        locker, nearest = await self._get_locker()
        self.is_taken = False
        if self.lock_key == locker:
            _log.debug("Lock acquired!")
            # We own the lock, yay!
            self.is_taken = True
            return True
        else:
            self.is_taken = False
            if not blocking:
                return False
            # Let's look for the lock
            watch_key = nearest.key
            _log.debug("Lock not acquired, now watching %s", watch_key)
            while True:
                try:
                    r = await self.client.watch(watch_key, index=nearest.modifiedIndex + 1)
                    _log.debug("Detected variation for %s: %s", r.key, r.action)
                    return (await self._acquired(blocking=True))
                except aio_etcd.EtcdKeyNotFound:
                    _log.debug("Key %s not present anymore, moving on", watch_key)
                    return (await self._acquired(blocking=True))
                except etcd.EtcdLockExpired as e:
                    raise e
                except etcd.EtcdException:
                    _log.exception("Unexpected exception")

    @property
    def lock_key(self):
        if not self._sequence:
            raise ValueError("No sequence present.")
        return self.path + '/' + str(self._sequence)

    def _set_sequence(self, key):
        self._sequence = key.replace(self.path, '').lstrip('/')

    async def _find_lock(self):
        if self._sequence:
            try:
                res = await self.client.read(self.lock_key)
                self._uuid = res.value
                return True
            except aio_etcd.EtcdKeyNotFound:
                return False
        elif self._uuid:
            try:
                for r in (await self.client.read(self.path, recursive=True)).leaves:
                    if r.value == self._uuid:
                        self._set_sequence(r.key)
                        return True
            except aio_etcd.EtcdKeyNotFound:
                pass
        return False

    async def _get_locker(self):
        results = [res for res in
                   (await self.client.read(self.path, recursive=True)).leaves]
        if not self._sequence:
            await self._find_lock()
        l = sorted([r.key for r in results])
        _log.debug("Lock keys found: %s", l)
        try:
            i = l.index(self.lock_key)
            if i == 0:
                _log.debug("No key before our one, we are the locker")
                return (l[0], None)
            else:
                _log.debug("Locker: %s, key to watch: %s", l[0], l[i-1])
                return (l[0], next(x for x in results if x.key == l[i-1]))
        except ValueError as exc:
            # Something very wrong is going on, most probably
            # our lock has expired
            raise aio_etcd.EtcdLockExpired(u"Lock not found") from exc



import json

import logging
import aio_etcd as etcd

_log = logging.getLogger(__name__)


class EtcdAuthBase(object):
    entity = 'example'

    def __init__(self, client, name):
        self.client = client
        self.name = name
        self.uri = "{}/auth/{}s/{}".format(self.client.version_prefix,
                                           self.entity, self.name)

    async def get_names(self):
        key = "{}s".format(self.entity)
        uri = "{}/auth/{}".format(self.client.version_prefix, key)
        response = await self.client.api_execute(uri, self.client._MGET)
        res = json.loads((await response.read()).decode('utf-8'))[key]
        res = list(x[self.entity] if isinstance(x,dict) else x for x in res)
        return res

    async def read(self):
        try:
            response = await self.client.api_execute(self.uri, self.client._MGET)
        except etcd.EtcdInsufficientPermissions as e:
            _log.error("Any action on the authorization requires the root role")
            raise
        except etcd.EtcdKeyNotFound:
            _log.info("%s '%s' not found", self.entity, self.name)
            raise
        except Exception as e:
            _log.error("Failed to fetch %s in %s%s: %r",
                       self.entity, self.client._base_uri,
                       self.client.version_prefix, e)
            raise etcd.EtcdException(
                "Could not fetch {} '{}'".format(self.entity, self.name)) from e

        self._from_net((await response.read()))

    async def write(self):
        try:
            r = self.__class__(self.client, self.name)
            await r.read()
        except etcd.EtcdKeyNotFound:
            r = None
        try:
            for payload in self._to_net(r):
                response = await self.client.api_execute_json(self.uri,
                                                        self.client._MPUT,
                                                        params=payload)
                # This will fail if the response is an error
                self._from_net(await response.read())
        except etcd.EtcdInsufficientPermissions as e:
            _log.error("Any action on the authorization requires the root role")
            raise
        except Exception as e:
            _log.error("Failed to write %s '%s'", self.entity, self.name)
            # TODO: fine-grained exception handling
            raise
            raise etcd.EtcdException(
                "Could not write {} '{}': {}".format(self.entity,
                                                     self.name, e)) from e

    async def delete(self):
        try:
            await self.client.api_execute(self.uri, self.client._MDELETE)
        except etcd.EtcdInsufficientPermissions as e:
            _log.error("Any action on the authorization requires the root role")
            raise
        except etcd.EtcdKeyNotFound:
            _log.info("%s '%s' not found", self.entity, self.name)
            raise
        except Exception as e:
            _log.error("Failed to delete %s in %s%s: %r",
                       self.entity, self._base_uri, self.version_prefix, e)
            raise etcd.EtcdException(
                "Could not delete {} '{}'".format(self.entity, self.name)) from e

    def _from_net(self, data):
        raise NotImplementedError()

    def _to_net(self, old=None):
        raise NotImplementedError()

    @classmethod
    def new(cls, client, data):
        c = cls(client, data[cls.entity])
        c._from_net(data)
        return c


class EtcdUser(EtcdAuthBase):
    """Class to manage in a orm-like way etcd users"""
    entity = 'user'

    def __init__(self, client, name):
        super(EtcdUser, self).__init__(client, name)
        self._roles = set()
        self._password = None

    def _from_net(self, data):
        d = json.loads(data.decode('utf-8'))
        r = d.get('roles', [])
        r = list(k['role'] if isinstance(k,dict) else k for k in r)
        self.roles = r
        self.name = d.get('user')

    def _to_net(self, prevobj=None):
        if prevobj is None:
            retval = [{"user": self.name, "password": self._password,
                       "roles": list(self.roles)}]
        else:
            retval = []
            if self._password:
                retval.append({"user": self.name, "password": self._password})
            to_grant = list(self.roles - prevobj.roles)
            to_revoke = list(prevobj.roles - self.roles)
            if to_grant:
                retval.append({"user": self.name, "grant": to_grant})
            if to_revoke:
                retval.append({"user": self.name, "revoke": to_revoke})
        # Let's blank the password now
        # Even if the user can't be written we don't want it to leak anymore.
        self._password = None
        return retval

    @property
    def roles(self):
        return self._roles

    @roles.setter
    def roles(self, val):
        self._roles = set(val)

    @property
    def password(self):
        """Empty property for password."""
        return None

    @password.setter
    def password(self, new_password):
        """Change user's password."""
        self._password = new_password

    def __str__(self):
        return json.dumps(self._to_net()[0])



class EtcdRole(EtcdAuthBase):
    entity = 'role'

    def __init__(self, client, name):
        super(EtcdRole, self).__init__(client, name)
        self._read_paths = set()
        self._write_paths = set()

    def _from_net(self, data):
        d = json.loads(data.decode('utf-8'))
        self.name = d.get('role')

        try:
            kv = d["permissions"]["kv"]
        except:
            self._read_paths = set()
            self._write_paths = set()
            return

        self._read_paths = set(kv.get('read', []))
        self._write_paths = set(kv.get('write', []))

    def _to_net(self, prevobj=None):
        retval = []
        if prevobj is None:
            retval.append({
                "role": self.name,
                "permissions":
                {
                    "kv":
                    {
                        "read": list(self._read_paths),
                        "write": list(self._write_paths)
                    }
                }
            })
        else:
            to_grant = {
                'read': list(self._read_paths - prevobj._read_paths),
                'write': list(self._write_paths - prevobj._write_paths)
            }
            to_revoke = {
                'read': list(prevobj._read_paths - self._read_paths),
                'write': list(prevobj._write_paths - self._write_paths)
            }
            if [path for sublist in to_revoke.values() for path in sublist]:
                retval.append({'role': self.name, 'revoke': {'kv': to_revoke}})
            if [path for sublist in to_grant.values() for path in sublist]:
                retval.append({'role': self.name, 'grant': {'kv': to_grant}})
        return retval

    def grant(self, path, permission):
        if 'R' in permission:
            self._read_paths.add(path)
        if 'W' in permission:
            self._write_paths.add(path)

    def revoke(self, path, permission):
        if 'R' in permission and \
           path in self._read_paths:
            self._read_paths.remove(path)
        if 'W' in permission and \
           path in self._write_paths:
            self._write_paths.remove(path)

    @property
    def acls(self):
        perms = {}
        try:
            for path in self._read_paths:
                perms[path] = 'R'
            for path in self._write_paths:
                if path in perms:
                    perms[path] += 'W'
                else:
                    perms[path] = 'W'
        except:
            pass
        return perms

    @acls.setter
    def acls(self, acls):
        self._read_paths = set()
        self._write_paths = set()
        for path, permission in acls.items():
            self.grant(path, permission)

    def __str__(self):
        return json.dumps({"role": self.name, 'acls': self.acls})


class Auth(object):
    def __init__(self, client):
        self.client = client
        self.uri = "{}/auth/enable".format(self.client.version_prefix)

    async def get_active(self):
        resp = await self.client.api_execute(self.uri, self.client._MGET)
        return json.loads((await resp.read()).decode('utf-8'))['enabled']

    async def set_active(self, value):
        if value != (await self.get_active()):
            method = value and self.client._MPUT or self.client._MDELETE
            await self.client.api_execute(self.uri, method)



#!/usr/bin/python3

from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()

def get_version(fname='aio_etcd/__init__.py'):
    with open(fname) as f:
        for line in f:
            if line.startswith('__VERSION__'):
                return eval(line.split('=')[-1])

version = '0.4.3'

install_requires = [
    'python-etcd',
    'aiohttp',
    'dnspython3',
    'urllib3>=1.7.1',
]

test_requires = [
    'pytest',
]

from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['--assert=plain','tests/']

#    def finalize_options(self):
#        TestCommand.finalize_options(self)
#        self.test_args = []
#        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    name='aio_etcd',
    version='.'.join(str(x) for x in get_version()),
    description="An asynchronous python client for etcd",
    long_description=README,
    classifiers=[
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Database :: Front-Ends",
    ],
    keywords='etcd raft distributed log api client',
    author='Matthias Urlichs',
    author_email='matthias@urlichs.de',
    url='http://github.com/M-o-a-T/aio-etcd',
    license='MIT',
    packages=('aio_etcd',),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=test_requires,
    test_suite='pytest.collector',
    cmdclass = {'test': PyTest},
)
